/*
 * Copyright notice from original mutt:
 * Copyright (C) 2004 Thomas Glanzmann <sithglan@stud.uni-erlangen.de>
 * Copyright (C) 2004 Tobias Werth <sitowert@stud.uni-erlangen.de>
 * Copyright (C) 2004 Brian Fundakowski Feldman <green@FreeBSD.org>
 *
 * This file is part of mutt-ng, see http://www.muttng.org/.
 * It's licensed under the GNU General Public License,
 * please see the file GPL in the top level source directory.
 */

#include <lib-lib/lib-lib.h>

#ifdef USE_HCACHE

#if defined(HAVE_QDBM)
#include <depot.h>
#include <cabin.h>
#include <villa.h>
#elif defined(HAVE_GDBM)
#include <gdbm.h>
#endif

#include <imap/message.h>

#include "charset.h"
#include "mutt.h"
#include "hcache.h"

struct hcache_t {
#if defined(HAVE_QDBM)
    VILLA *db;
#elif defined(HAVE_GDBM)
    GDBM_FILE db;
#endif
    char *folder;
    unsigned int crc;
};

typedef union {
  struct timeval timeval;
  unsigned long uid_validity;
} validate;

#define UPPER4K(i)  ((i & ~(4096 - 1)) + 4096)

static unsigned char *lazy_malloc(ssize_t siz)
{
    return p_new(unsigned char, UPPER4K(siz));
}

static void lazy_realloc(unsigned char **p, ssize_t siz)
{
    p_realloc(p, UPPER4K(siz));
}

static unsigned char *dump_int (unsigned int i, unsigned char *d, int *off)
{
  lazy_realloc (&d, *off + sizeof (int));
  memcpy (d + *off, &i, sizeof (int));
  (*off) += sizeof (int);

  return d;
}

static void restore_int (unsigned int *i, const unsigned char *d, int *off)
{
  memcpy (i, d + *off, sizeof (int));
  (*off) += sizeof (int);
}

static unsigned char *dump_char (char *c, unsigned char *d, int *off)
{
  unsigned int size;

  if (c == NULL) {
    size = 0;
    d = dump_int (size, d, off);
    return d;
  }

  size = m_strlen(c) + 1;
  d = dump_int (size, d, off);
  lazy_realloc (&d, *off + size);
  memcpy (d + *off, c, size);
  *off += size;

  return d;
}

static void restore_char (char **c, const unsigned char *d, int *off)
{
  unsigned int size;

  restore_int (&size, d, off);

  if (size == 0) {
    *c = NULL;
    return;
  }

  *c = p_dup(d + *off, size);
  *off += size;
}

static unsigned char *dump_address (address_t * a, unsigned char *d, int *off)
{
  unsigned int counter = 0;
  unsigned int start_off = *off;

  d = dump_int (0xdeadbeef, d, off);

  while (a) {
    d = dump_char (a->personal, d, off);
    d = dump_char (a->mailbox, d, off);
    d = dump_int (a->group, d, off);
    a = a->next;
    counter++;
  }

  memcpy (d + start_off, &counter, sizeof (int));

  return d;
}

static void restore_address (address_t ** a, const unsigned char *d, int *off)
{
  unsigned int counter;

  restore_int (&counter, d, off);

  while (counter) {
    *a = p_new(address_t, 1);
    restore_char (&(*a)->personal, d, off);
    restore_char (&(*a)->mailbox, d, off);
    restore_int ((unsigned int *) &(*a)->group, d, off);
    a = &(*a)->next;
    counter--;
  }

  *a = NULL;
}

static unsigned char *dump_list (string_list_t * l, unsigned char *d, int *off)
{
  unsigned int counter = 0;
  unsigned int start_off = *off;

  d = dump_int (0xdeadbeef, d, off);

  while (l) {
    d = dump_char (l->data, d, off);
    l = l->next;
    counter++;
  }

  memcpy (d + start_off, &counter, sizeof (int));

  return d;
}

static void restore_list (string_list_t ** l, const unsigned char *d, int *off)
{
  unsigned int counter;

  restore_int (&counter, d, off);

  while (counter) {
    *l = p_new(string_list_t, 1);
    restore_char (&(*l)->data, d, off);
    l = &(*l)->next;
    counter--;
  }

  *l = NULL;
}

static unsigned char *dump_parameter (parameter_t * p, unsigned char *d,
                                      int *off)
{
  unsigned int counter = 0;
  unsigned int start_off = *off;

  d = dump_int (0xdeadbeef, d, off);

  while (p) {
    d = dump_char (p->attribute, d, off);
    d = dump_char (p->value, d, off);
    p = p->next;
    counter++;
  }

  memcpy (d + start_off, &counter, sizeof (int));

  return d;
}

static void
restore_parameter (parameter_t ** p, const unsigned char *d, int *off)
{
  unsigned int counter;

  restore_int (&counter, d, off);

  while (counter) {
    *p = parameter_new();
    restore_char (&(*p)->attribute, d, off);
    restore_char (&(*p)->value, d, off);
    p = &(*p)->next;
    counter--;
  }
}

static unsigned char *dump_body (BODY * c, unsigned char *d, int *off)
{
  lazy_realloc (&d, *off + sizeof (BODY));
  memcpy (d + *off, c, sizeof (BODY));
  *off += sizeof (BODY);

  d = dump_char (c->xtype, d, off);
  d = dump_char (c->subtype, d, off);

  d = dump_parameter (c->parameter, d, off);

  d = dump_char (c->description, d, off);
  d = dump_char (c->form_name, d, off);
  d = dump_char (c->filename, d, off);
  d = dump_char (c->d_filename, d, off);

  return d;
}

static void restore_body (BODY * c, const unsigned char *d, int *off)
{
  memcpy (c, d + *off, sizeof (BODY));
  *off += sizeof (BODY);

  restore_char (&c->xtype, d, off);
  restore_char (&c->subtype, d, off);

  restore_parameter (&c->parameter, d, off);

  restore_char (&c->description, d, off);
  restore_char (&c->form_name, d, off);
  restore_char (&c->filename, d, off);
  restore_char (&c->d_filename, d, off);
}

static unsigned char *dump_envelope (ENVELOPE * e, unsigned char *d, int *off)
{
  d = dump_address (e->return_path, d, off);
  d = dump_address (e->from, d, off);
  d = dump_address (e->to, d, off);
  d = dump_address (e->cc, d, off);
  d = dump_address (e->bcc, d, off);
  d = dump_address (e->sender, d, off);
  d = dump_address (e->reply_to, d, off);
  d = dump_address (e->mail_followup_to, d, off);

  d = dump_char (e->subject, d, off);
  if (e->real_subj) {
    d = dump_int (e->real_subj - e->subject, d, off);
  }
  else {
    d = dump_int (-1, d, off);
  }
  d = dump_char (e->message_id, d, off);
  d = dump_char (e->supersedes, d, off);
  d = dump_char (e->date, d, off);
  d = dump_char (e->x_label, d, off);
  d = dump_char (e->list_post, d, off);

#ifdef USE_NNTP
  d = dump_char (e->newsgroups, d, off);
  d = dump_char (e->xref, d, off);
  d = dump_char (e->followup_to, d, off);
  d = dump_char (e->x_comment_to, d, off);
#endif

  d = dump_list (e->references, d, off);
  d = dump_list (e->in_reply_to, d, off);
  d = dump_list (e->userhdrs, d, off);

  return d;
}

static void restore_envelope (ENVELOPE * e, const unsigned char *d, int *off)
{
  int real_subj_off;

  restore_address (&e->return_path, d, off);
  restore_address (&e->from, d, off);
  restore_address (&e->to, d, off);
  restore_address (&e->cc, d, off);
  restore_address (&e->bcc, d, off);
  restore_address (&e->sender, d, off);
  restore_address (&e->reply_to, d, off);
  restore_address (&e->mail_followup_to, d, off);

  restore_char (&e->subject, d, off);
  restore_int ((unsigned int *) (&real_subj_off), d, off);
  if (0 <= real_subj_off) {
    e->real_subj = e->subject + real_subj_off;
  }
  else {
    e->real_subj = NULL;
  }
  restore_char (&e->message_id, d, off);
  restore_char (&e->supersedes, d, off);
  restore_char (&e->date, d, off);
  restore_char (&e->x_label, d, off);
  restore_char (&e->list_post, d, off);

#ifdef USE_NNTP
  restore_char (&e->newsgroups, d, off);
  restore_char (&e->xref, d, off);
  restore_char (&e->followup_to, d, off);
  restore_char (&e->x_comment_to, d, off);
#endif

  restore_list (&e->references, d, off);
  restore_list (&e->in_reply_to, d, off);
  restore_list (&e->userhdrs, d, off);
}

static unsigned int crc32(unsigned int crc, const void *src, ssize_t len)
{
    int i;
    const unsigned char *p = src;

    while (len--) {
        crc ^= *p++;
        for (i = 0; i < 8; i++)
            crc = (crc >> 1) ^ ((crc & 1) ? 0xedb88320 : 0);
    }
    return crc;
}

static int generate_crc32(void)
{
    static int crc = 0;

    crc = crc32(crc, "madmutt.2007.05.13", m_strlen("madmutt.2007.05.13"));
#ifdef HAVE_LANGINFO_CODESET
    crc = crc32(crc, MCharset.charset, m_strlen(MCharset.charset));
    crc = crc32(crc, "HAVE_LANGINFO_CODESET",
                m_strlen("HAVE_LANGINFO_CODESET"));
#endif
    crc = crc32(crc, "USE_POP",   m_strlen("USE_POP"));
    crc = crc32(crc, "MIXMASTER", m_strlen("MIXMASTER"));
    crc = crc32(crc, "USE_IMAP",  m_strlen("USE_IMAP"));
#ifdef USE_NNTP
    crc = crc32(crc, "USE_NNTP",  m_strlen("USE_NNTP"));
#endif
    return crc;
}

static const char *
mutt_hcache_per_folder(const char *path, const char *folder)
{
    static char buf[_POSIX_PATH_MAX];
    struct stat st;
    int pos;

    if (stat(path, &st) < 0 || !S_ISDIR(st.st_mode)) {
        return path;
    }

    pos  = m_strcpy(buf, sizeof(buf), path);
    pos += m_strputc(buf + pos, sizeof(buf) - pos, '/');

    while (*folder) {
        if (isalnum((unsigned char)*folder) || *folder == '.') {
            pos += m_strputc(buf + pos, sizeof(buf) - pos, *folder);
        } else {
            pos += m_strputc(buf + pos, sizeof(buf) - pos, '_');
        }
        folder++;
    }
    pos += m_strcpy(buf + pos, sizeof(buf) - pos, ".hdb");
    return buf;
}

/* This function transforms a header into a char so that it is useable by
 * db_store */
static void *mutt_hcache_dump (void *_db, HEADER * h, int *off,
                               unsigned long uid_validity)
{
  struct hcache_t *db = _db;
  unsigned char *d = NULL;

  *off = 0;

  d = lazy_malloc (sizeof (validate));

  if (uid_validity) {
    memcpy (d, &uid_validity, sizeof (unsigned long));
  }
  else {
    struct timeval now;

    gettimeofday (&now, NULL);
    memcpy (d, &now, sizeof (struct timeval));
  }
  *off += sizeof (validate);

  d = dump_int (db->crc, d, off);

  lazy_realloc (&d, *off + sizeof (HEADER));
  memcpy (d + *off, h, sizeof (HEADER));
  *off += sizeof (HEADER);

  d = dump_envelope (h->env, d, off);
  d = dump_body (h->content, d, off);
  d = dump_char (h->maildir_flags, d, off);

  return d;
}

HEADER *mutt_hcache_restore (const unsigned char *d, HEADER ** oh)
{
  int off = 0;
  HEADER *h = header_new();

  /* skip validate */
  off += sizeof (validate);

  /* skip crc */
  off += sizeof (unsigned int);

  memcpy (h, d + off, sizeof (HEADER));
  off += sizeof (HEADER);

  h->env = envelope_new();
  restore_envelope (h->env, d, &off);

  h->content = body_new();
  restore_body (h->content, d, &off);

  restore_char (&h->maildir_flags, d, &off);

  /* this is needed for maildir style mailboxes */
  if (oh) {
    h->old = (*oh)->old;
    h->path = m_strdup((*oh)->path);
    header_delete(oh);
  }

  return h;
}

hcache_t *mutt_hcache_open(const char *path, const char *folder)
{
    hcache_t *h = p_new(hcache_t, 1);

    h->folder = m_strdup(folder);
    h->crc = generate_crc32();

    if (m_strisempty(path)) {
        p_delete(&h->folder);
        p_delete(&h);
        return NULL;
    }

    path = mutt_hcache_per_folder(path, folder);

    {
#if defined(HAVE_QDBM)
        int flags = VL_OWRITER | VL_OCREAT;
        if (option(OPTHCACHECOMPRESS))
            flags |= VL_OZCOMP;

        h->db = vlopen(path, flags, VL_CMPLEX);
#elif defined(HAVE_GDBM)
        int pagesize = atoi(HeaderCachePageSize) ?: 16384;

        h->db = gdbm_open((char *) path, pagesize, GDBM_WRCREAT, 00600, NULL);
#endif
    }

    if (!h->db) {
        p_delete(&h->folder);
        p_delete(&h);
    }
    return h;
}

void mutt_hcache_close(hcache_t **db)
{
    if (!*db)
        return;

#if defined(HAVE_QDBM)
    vlclose((*db)->db);
#elif defined(HAVE_GDBM)
    gdbm_close((*db)->db);
#endif

    p_delete(&(*db)->folder);
    p_delete(db);
}

void *mutt_hcache_fetch(hcache_t *db, const char *filename,
                        ssize_t (*keylen)(const char *fn))
{
    char path[_POSIX_PATH_MAX];
    void *data = NULL;

    if (!db)
        return NULL;

    snprintf(path, sizeof(path), "%s%s", db->folder, filename);

    {
#if defined(HAVE_QDBM)
        int ksize = strlen(db->folder) + keylen(path + strlen(db->folder));
        data  = vlget(db->db, path, ksize, NULL);
#elif defined(HAVE_GDBM)
        datum k = { .dptr = path, .dsize = keylen(path) };

        data = gdbm_fetch(db->db, k).dtpr;
#endif
    }

    if (data) {
        int off = sizeof(validate);
        unsigned crc = 0;

        restore_int(&crc, data, &off);
        if (crc != db->crc)
            p_delete(&data);
    }

    return data;
}

int mutt_hcache_store(hcache_t *db, const char *filename, HEADER *header,
                      unsigned long uid_validity,
                      ssize_t (*keylen)(const char *fn))
{
    char path[_POSIX_PATH_MAX];
    int ret, dsize;
    void *data = NULL;

    if (!db)
        return -1;

    snprintf(path, sizeof(path), "%s%s", db->folder, filename);
    data = mutt_hcache_dump(db, header, &dsize, uid_validity);

    {
#if defined(HAVE_QDBM)
        int ksize = strlen(db->folder) + keylen(path + strlen(db->folder));

        ret = vlput(db->db, path, ksize, data, dsize, VL_DOVER);
#elif defined(HAVE_GDBM)
        datum k = { .dptr = path, .dsize = keylen(path) };
        datum v = { .dptr = data, .dsize = dsize };

        ret = gdbm_store (db->db, k, v, GDBM_REPLACE);
#endif
    }

    p_delete(&data);
    return ret;
}

int mutt_hcache_delete(hcache_t *db, const char *filename,
                       ssize_t (*keylen)(const char *fn))
{
    char path[_POSIX_PATH_MAX];

    if (!db)
        return -1;

    snprintf(path, sizeof(path), "%s%s", db->folder, filename);

    {
#if defined(HAVE_QDBM)
        int ksize = strlen(db->folder) + keylen(path + strlen(db->folder));
        return vlout(db->db, path, ksize);
#elif defined(HAVE_GDBM)
        datum k = { .dptr = path, .dsize = keylen(path) };
        return gdbm_delete(db->db, k);
#endif
    }
}

#endif /* USE_HCACHE */
